import os
from argparse import ArgumentParser
from configparser import ConfigParser, MissingSectionHeaderError
from distutils.util import strtobool


def read_from_conf(
    conf_file, conf_section='defaults', conf_key_map=None, no_strip_quote=False
):
    """Read key/value from conf_section of conf_file.
    Hyphens (-) in keys will be replace with underscores (_).
    All keys and values are considered as strings.

    Use update_parser_defaults_with_conf (2nd return value)
    instead to get result with values converted to correct types.

    Args:
        conf_file:
            INI-like conf file to find defaults key/value.
        conf_section:
            Section in conf_file.
            If section doesn't exist then make one and set as default.
        conf_key_map:
            Mapping of keys parsed from conf file.
            This is useful if you want to replace an old key name with a new one.
            e.g. to make your code backward compatible when you want to
            change parameter's name.
        no_strip_quote:
            Do not strip single/double quotes from values in conf_file.

    Returns:
        Dict of key/value in configuration file.
    """
    conf_file = os.path.expanduser(conf_file)
    if not os.path.exists(conf_file):
        raise FileNotFoundError('conf_file does not exist. f={f}'.format(f=conf_file))

    config = ConfigParser()
    with open(conf_file) as fp:
        s = fp.read()
        try:
            config.read_string(s)
        except MissingSectionHeaderError:
            section = '[{sect}]\n'.format(sect=conf_section)
            config.read_string(section + s)

    d_ = dict(config.items(conf_section))
    result = {}
    for k, v in d_.items():
        if not no_strip_quote:
            v = v.strip('"\'')
        if v:
            k_ = k.replace('-', '_')
            if conf_key_map and k_ in conf_key_map:
                k_ = conf_key_map[k_]
            result[k_] = v

    return result


def update_parsers_defaults_with_conf(
    parsers,
    conf_file,
    conf_section='defaults',
    conf_key_map=None,
    no_strip_quote=False,
    val_type=None,
    val_default=None,
):
    """Updates argparse.ArgumentParser's defaults with key/value pairs
    defined in conf_file. Also, returns a dict of key/values defined in
    conf_file with correct type for each value.

    Type of each value in conf_file can be guessed from:
        - default value of ArgumentParser's argument.
        - val_type if it's given.
        - val_default if it's given.
    Otherwise it is considered as string type since ArgumentParser
    does not allow direct access to each argument's type.
    Therefore, this function tries to best-guess such type.

    This function does not work recursively with subparsers.
    Therefore, call this function with each subparser to update each
    subparser's defaults.

    Args:
        parsers:
            List of argparse.ArgumentParser objects to be updated with
            new defaults defined in conf_file. Useful for subparsers.
        conf_file:
            See read_from_conf()
        conf_section:
            See read_from_conf()
        conf_key_map:
            See read_from_conf()
        no_strip_quote:
            See read_from_conf()
        val_type:
            {key: value's type} where key is a key in conf_file.
            If not defined, var's type can be guessed either from
            parser's default value or from val_default.
            parser's default will override all otehrs.
        val_default:
            {key: value's default} where key is a key in conf_file.
            Type can be guessed from argument's default value.

    Returns:
        Dict of key/value pair parsed from conf_file.
        Such value is converted into a correct type which is
        guessed from val_type, val_default and arguments' defaults
        defined in parsers.
    """
    if isinstance(parsers, ArgumentParser):
        parsers = [parsers]

    defaults = read_from_conf(
        conf_file=conf_file,
        conf_section=conf_section,
        conf_key_map=conf_key_map,
        no_strip_quote=no_strip_quote,
    )

    if val_default:
        for k, v in val_default.items():
            if k not in defaults:
                defaults[k] = None

    # used "is not None" for guessed_default to catch boolean false
    for k, v in defaults.items():
        if val_default and k in val_default:
            guessed_default = val_default[k]
        else:
            for p in parsers:
                guessed_default = p.get_default(k)
                if guessed_default:
                    break
        if val_type and k in val_type:
            guessed_type = val_type[k]
        elif guessed_default is not None:
            guessed_type = type(guessed_default)
        else:
            guessed_type = None

        if v is None and guessed_default is not None:
            v = guessed_default
            defaults[k] = v

        if guessed_type:
            if guessed_type is bool and isinstance(v, str):
                defaults[k] = bool(strtobool(v))
            else:
                defaults[k] = guessed_type(v)

    # update ArgumentParser's default and then return defaults dict
    for p in parsers:
        p.set_defaults(**defaults)
    return defaults



import argparse
import os
from enum import Enum

from autouri import URIBase

from .arg_tool import update_parsers_defaults_with_conf
from .backward_compatibility import PARAM_KEY_NAME_CHANGE
from .caper_workflow_opts import CaperWorkflowOpts
from .cromwell import Cromwell
from .cromwell_backend import (
    CALL_CACHING_DUP_STRAT_COPY,
    CALL_CACHING_DUP_STRAT_REFERENCE,
    LOCAL_HASH_STRAT_FILE,
    LOCAL_HASH_STRAT_PATH,
    LOCAL_HASH_STRAT_PATH_MTIME,
    CromwellBackendAWS,
    CromwellBackendBase,
    CromwellBackendCommon,
    CromwellBackendDatabase,
    CromwellBackendGCP,
    CromwellBackendLocal,
)
from .cromwell_rest_api import CromwellRestAPI
from .resource_analysis import ResourceAnalysis
from .server_heartbeat import ServerHeartbeat
from .singularity import Singularity

DEFAULT_CAPER_CONF = '~/.caper/default.conf'
DEFAULT_LIST_FORMAT = 'id,status,name,str_label,user,parent,submission'
DEFAULT_OUT_DIR = '.'
DEFAULT_CROMWELL_STDOUT = './cromwell.out'


class ResourceAnalysisReductionMethod(Enum):
    sum = sum
    max = max
    min = min
    none = None


def get_defaults(conf_file=None):
    """Wrapper for `get_parser_and_defaults()`.
    Use this function to get default values updated with `conf_file`.

    Args:
        conf_file:
            `DEFAULT_CAPER_CONF` will be used if it is None.

    Returns updated defaults only.
    """
    _, conf_dict = get_parser_and_defaults(conf_file=conf_file)
    return conf_dict


def get_parser_and_defaults(conf_file=None):
    """Creates a main parser and make a subparser for each subcommand.
    There are many parent parsers defined here.
    Each subparser will take a certain combination of these parent parsers
    to share the same parameter arguments between subcommands.
    e.g. subcommand run and server share the same --cromwell argument, which
    is defined in a parent parser "parent_runner".

    Finally each subparser's default is updated with values defined in conf_file.

    Args:
        conf_file:
            If defined, this will be used instead of partially parsing command line
            arguments to find conf_file (-c).
            `DEFAULT_CAPER_CONF` will be used if it is None.
    Returns:
        parser:
            ArgumentParser object with all arguments defined for each sub-
            command (subparser).
        conf_dict:
            Dict of key/value pairs parsed from conf_file.
            Such value is converted into a correct type guessed from
            defaults of arguments defined in ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        description='Caper (Cromwell-assisted Pipeline ExecutioneR)'
    )
    parser.add_argument('-v', '--version', action='store_true', help='Show version')

    subparser = parser.add_subparsers(dest='action')

    parent_init = argparse.ArgumentParser(add_help=False)
    parent_init.add_argument('platform', help='Platform to initialize Caper for.')

    # all
    parent_all = argparse.ArgumentParser(add_help=False)
    parent_all.add_argument(
        '-c', '--conf', help='Specify config file', default=DEFAULT_CAPER_CONF
    )
    parent_all.add_argument(
        '-D', '--debug', action='store_true', help='Prints all logs >= DEBUG level'
    )
    parent_all.add_argument(
        '--gcp-service-account-key-json',
        help='Secret key JSON file for Google Cloud Platform service account. '
        'This service account should have enough permission to '
        'Storage for client functions and '
        'Storage/Compute Engine/Genomics API/Life Sciences API '
        'for server/runner functions.',
    )

    group_loc = parent_all.add_argument_group(
        title='cache directories for localization'
    )
    group_loc.add_argument(
        '--local-loc-dir',
        '--tmp-dir',
        help='Temporary directory to store Cromwell\'s intermediate backend files. '
        'These files include backend.conf, workflow_opts.json, imports.zip. and '
        'localized input JSON files due to deepcopying (recursive localization). '
        'Cromwell\'s MySQL/PostgreSQL DB password can be exposed on backend.conf '
        'on this directory. Therefore, DO NOT USE /tmp HERE. This directory is '
        'also used for storing cached files for local/slurm/sge/pbs backends.',
    )
    group_loc.add_argument(
        '--gcp-loc-dir',
        '--tmp-gcs-bucket',
        help='Temporary directory to store cached files for gcp backend. '
        'e.g. gs://my-bucket/caper-cache-dir. ',
    )
    group_loc.add_argument(
        '--aws-loc-dir',
        '--tmp-s3-bucket',
        help='Temporary directory to store cached files for aws backend. '
        'e.g. s3://my-bucket/caper-cache-dir. ',
    )

    # run, server, submit
    parent_backend = argparse.ArgumentParser(add_help=False)
    parent_backend.add_argument('-b', '--backend', help='Backend to run a workflow')
    parent_backend.add_argument(
        '--dry-run',
        action='store_true',
        help='Caper localizes remote files and validates WDL '
        'but does not run/submit a pipeline.',
    )

    # run, server
    parent_runner = argparse.ArgumentParser(add_help=False)

    parent_runner = parent_runner.add_argument_group(title='Cromwell logging arguments')
    parent_runner.add_argument(
        '--cromwell-stdout',
        default=DEFAULT_CROMWELL_STDOUT,
        help='Local file to write STDOUT of Cromwell Java process to. '
        'This is for Cromwell (not for Caper\'s logging system).',
    )
    group_db = parent_runner.add_argument_group(
        title='General DB settings (for both file DB and MySQL DB)'
    )
    group_db.add_argument(
        '--db',
        default=CromwellBackendDatabase.DEFAULT_DB,
        help='Cromwell metadata database type',
    )
    group_db.add_argument(
        '--db-timeout',
        type=int,
        default=CromwellBackendDatabase.DEFAULT_DB_TIMEOUT_MS,
        help='Milliseconds to wait for DB connection.',
    )

    group_file_db = parent_runner.add_argument_group(
        title='HyperSQL file DB arguments (unstable, not recommended)'
    )
    group_file_db.add_argument(
        '--file-db',
        '-d',
        help='Default DB file for Cromwell\'s built-in HyperSQL database.',
    )

    group_mysql = parent_runner.add_argument_group(title='MySQL DB arguments')
    group_mysql.add_argument(
        '--mysql-db-ip',
        default=CromwellBackendDatabase.DEFAULT_MYSQL_DB_IP,
        help='MySQL Database IP address (e.g. localhost)',
    )
    group_mysql.add_argument(
        '--mysql-db-port',
        type=int,
        default=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PORT,
        help='MySQL Database TCP/IP port (e.g. 3306)',
    )
    group_mysql.add_argument(
        '--mysql-db-user',
        default=CromwellBackendDatabase.DEFAULT_MYSQL_DB_USER,
        help='MySQL DB username',
    )
    group_mysql.add_argument(
        '--mysql-db-password',
        default=CromwellBackendDatabase.DEFAULT_MYSQL_DB_PASSWORD,
        help='MySQL DB password',
    )
    group_mysql.add_argument(
        '--mysql-db-name',
        default=CromwellBackendDatabase.DEFAULT_MYSQL_DB_NAME,
        help='MySQL DB name for Cromwell',
    )

    group_postgresql = parent_runner.add_argument_group(title='PostgreSQL DB arguments')
    group_postgresql.add_argument(
        '--postgresql-db-ip',
        default=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_IP,
        help='PostgreSQL DB IP address (e.g. localhost)',
    )
    group_postgresql.add_argument(
        '--postgresql-db-port',
        type=int,
        default=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PORT,
        help='PostgreSQL DB TCP/IP port (e.g. 5432)',
    )
    group_postgresql.add_argument(
        '--postgresql-db-user',
        default=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_USER,
        help='PostgreSQL DB username',
    )
    group_postgresql.add_argument(
        '--postgresql-db-password',
        default=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_PASSWORD,
        help='PostgreSQL DB password',
    )
    group_postgresql.add_argument(
        '--postgresql-db-name',
        default=CromwellBackendDatabase.DEFAULT_POSTGRESQL_DB_NAME,
        help='PostgreSQL DB name for Cromwell',
    )

    group_cromwell = parent_runner.add_argument_group(title='Cromwell settings')
    group_cromwell.add_argument(
        '--cromwell',
        default=Cromwell.DEFAULT_CROMWELL,
        help='Path or URL for Cromwell JAR file',
    )
    group_cromwell.add_argument(
        '--max-concurrent-tasks',
        type=int,
        default=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        help='Number of concurrent tasks. '
        '"config.concurrent-job-limit" in Cromwell backend configuration '
        'for each backend',
    )
    group_cromwell.add_argument(
        '--max-concurrent-workflows',
        type=int,
        default=CromwellBackendCommon.DEFAULT_MAX_CONCURRENT_WORKFLOWS,
        help='Number of concurrent workflows. '
        '"system.max-concurrent-workflows" in backend configuration',
    )
    group_cromwell.add_argument(
        '--memory-retry-error-keys',
        default=','.join(CromwellBackendCommon.DEFAULT_MEMORY_RETRY_ERROR_KEYS),
        help='(CURRENTLY NOT WORKING) '
        'If an error caught by these comma-separated keys occurs, '
        'then increase memory by --memory-retry-multiplier '
        'for retrials controlled by --max-retries. '
        'See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory/ '
        'for details. ',
    )
    group_cromwell.add_argument(
        '--disable-call-caching',
        action='store_true',
        help='Disable Cromwell\'s call caching, which re-uses outputs from '
        'previous workflows',
    )
    group_cromwell.add_argument(
        '--backend-file',
        help='Custom Cromwell backend configuration file to override all',
    )
    group_cromwell.add_argument(
        '--soft-glob-output',
        action='store_true',
        help='Use soft-linking when globbing outputs for a filesystem that '
        'does not allow hard-linking. e.g. beeGFS. '
        'This flag does not work with backends based on a Docker container. '
        'i.e. gcp and aws. Also, '
        'it does not work with local backends (local/slurm/sge/pbs) '
        'with --. However, it works fine with --singularity.',
    )
    group_cromwell.add_argument(
        '--local-hash-strat',
        default=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        choices=[
            LOCAL_HASH_STRAT_FILE,
            LOCAL_HASH_STRAT_PATH,
            LOCAL_HASH_STRAT_PATH_MTIME,
        ],
        help='File hashing strategy for call caching. '
        'For local backends (local/slurm/sge/pbs) only. '
        'file: use md5sum hash (slow), path: use path only, '
        'path+modtime (default): use path + mtime.',
    )

    group_local = parent_runner.add_argument_group(title='local backend arguments')
    group_local.add_argument(
        '--local-out-dir',
        '--out-dir',
        default=DEFAULT_OUT_DIR,
        help='Output directory path for local backend. '
        'Cloud backends (gcp, aws) use different output directories. '
        'For gcp, define --gcp-out-dir. '
        'For aws, define --aws-out-dir.',
    )

    group_gc_all = parent_backend.add_argument_group(
        title='GCP backend arguments for server/runner/client'
    )
    group_gc = parent_runner.add_argument_group(
        title='GCP backend arguments for server/runner'
    )
    group_gc.add_argument('--gcp-prj', help='GC project')
    group_gc_all.add_argument(
        '--use-google-cloud-life-sciences',
        action='store_true',
        help='Use Google Cloud Life Sciences API (v2beta) instead of '
        'deprecated Genomics API (v2alpha1).'
        'Life Sciences API requires only one region specified with'
        'gcp-region. gcp-zones will be ignored since it is for Genomics API.'
        'See https://cloud.google.com/life-sciences/docs/concepts/locations '
        'for supported regions.',
    )
    group_gc.add_argument(
        '--gcp-region',
        default=CromwellBackendGCP.DEFAULT_REGION,
        help='GCP region for Google Cloud Life Sciences API. '
        'This is used only when --use-google-cloud-life-sciences is defined.',
    )
    group_gc_all.add_argument(
        '--gcp-zones',
        help='Comma-separated GCP zones used for Genomics API. '
        '(e.g. us-west1-b,us-central1-b). '
        'If you use --use-google-cloud-life-sciences then '
        'define --gcp-region instead.',
    )
    group_gc.add_argument(
        '--gcp-call-caching-dup-strat',
        default=CromwellBackendGCP.DEFAULT_CALL_CACHING_DUP_STRAT,
        choices=[CALL_CACHING_DUP_STRAT_REFERENCE, CALL_CACHING_DUP_STRAT_COPY],
        help='Duplication strategy for call-cached outputs for GCP backend: '
        'copy: make a copy, reference: refer to old output in metadata.json.',
    )
    group_gc.add_argument(
        '--gcp-out-dir',
        '--out-gcs-bucket',
        help='Output directory path for GCP backend. ' 'e.g. gs://my-bucket/my-output.',
    )

    group_aws = parent_runner.add_argument_group(title='AWS backend arguments')
    group_aws.add_argument('--aws-batch-arn', help='ARN for AWS Batch')
    group_aws.add_argument('--aws-region', help='AWS region (e.g. us-west-1)')
    group_aws.add_argument(
        '--aws-out-dir',
        '--out-s3-bucket',
        help='Output path on S3 bucket for AWS backend. '
        'e.g. s3://my-bucket/my-output.',
    )
    group_aws.add_argument(
        '--aws-call-caching-dup-strat',
        default=CromwellBackendAWS.DEFAULT_CALL_CACHING_DUP_STRAT,
        choices=[CALL_CACHING_DUP_STRAT_REFERENCE, CALL_CACHING_DUP_STRAT_COPY],
        help='Duplication strategy for call-cached outputs for AWS backend: '
        'copy: make a copy, reference: refer to old output in metadata.json.',
    )

    # run, submit
    parent_submit = argparse.ArgumentParser(add_help=False)

    parent_submit.add_argument(
        'wdl',
        help='Path, URL or URI for WDL script '
        'Example: /scratch/my.wdl, gs://some/where/our.wdl, '
        'http://hello.com/world/your.wdl',
    )
    parent_submit.add_argument('-i', '--inputs', help='Workflow inputs JSON file')
    parent_submit.add_argument('-o', '--options', help='Workflow options JSON file')
    parent_submit.add_argument('-l', '--labels', help='Workflow labels JSON file')
    parent_submit.add_argument(
        '-p', '--imports', help='Zip file of imported subworkflows'
    )
    parent_submit.add_argument(
        '-s',
        '--str-label',
        help='Caper\'s special label for a workflow '
        'This label will be added to a workflow labels JSON file '
        'as a value for a key "caper-label". '
        'DO NOT USE IRREGULAR CHARACTERS. USE LETTERS, NUMBERS, '
        'DASHES AND UNDERSCORES ONLY (^[A-Za-z0-9\\-\\_]+$) '
        'since this label is used to compose a path for '
        'workflow\'s temporary/cache directory (.caper_tmp/label/timestamp/)',
    )
    parent_submit.add_argument(
        '--hold',
        action='store_true',
        help='Put a hold on a workflow when submitted to a Cromwell server.',
    )
    parent_submit.add_argument(
        '--singularity-cachedir',
        default=Singularity.DEFAULT_SINGULARITY_CACHEDIR,
        help='Singularity cache directory. Equivalent to exporting an '
        'environment variable SINGULARITY_CACHEDIR. '
        'Define it to prevent repeatedly building a singularity image '
        'for every pipeline task',
    )
    parent_submit.add_argument(
        '--use-gsutil-for-s3',
        action='store_true',
        help='Use gsutil CLI for direct trasnfer between S3 and GCS buckets. '
        'Otherwise, such file transfer will stream through local machine. '
        'Make sure that gsutil is installed on your system and it has access to '
        'credentials for AWS (e.g. ~/.boto or ~/.aws/credentials).',
    )
    parent_submit.add_argument(
        '--no-deepcopy',
        action='store_true',
        help='(IMPORTANT) --deepcopy has been deprecated. '
        'Deepcopying is now activated by default. '
        'This flag disables deepcopy for '
        'JSON (.json), TSV (.tsv) and CSV (.csv) '
        'files specified in an input JSON file (--inputs). '
        'Find all path/URI string values in an input JSON file '
        'and make copies of files on a local/remote storage '
        'for a target backend. Make sure that you have installed '
        'gsutil for GCS and aws for S3.',
    )
    parent_submit.add_argument(
        '--ignore-womtool',
        action='store_true',
        help='Ignore warnings from womtool.jar.',
    )
    parent_submit.add_argument(
        '--womtool',
        default=Cromwell.DEFAULT_WOMTOOL,
        help='Path or URL for Cromwell\'s womtool JAR file',
    )
    parent_submit.add_argument(
        '--java-heap-womtool',
        default=Cromwell.DEFAULT_JAVA_HEAP_WOMTOOL,
        help='Java heap size for Womtool (java -Xmx)',
    )
    parent_submit.add_argument(
        '--max-retries',
        type=int,
        default=CaperWorkflowOpts.DEFAULT_MAX_RETRIES,
        help='Number of retries for failing tasks. '
        'equivalent to "maxRetries" in workflow options JSON file.',
    )
    parent_submit.add_argument(
        '--memory-retry-multiplier',
        default=CaperWorkflowOpts.DEFAULT_MEMORY_RETRY_MULTIPLIER,
        help='(CURRENTLY NOT WORKING) '
        'If an error caught by --memory-retry-error-keys occurs, '
        'then increase memory by this '
        'for retrials controlled by --max-retries. '
        'See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory/ '
        'for details.',
    )
    parent_submit.add_argument(
        '--gcp-monitoring-script',
        default=CaperWorkflowOpts.DEFAULT_GCP_MONITORING_SCRIPT,
        help='Monitoring script for gcp backend only. '
        'Caper defaults to use its own monitoring script which works fine '
        'with subcommand gcp_profile. '
        'To make your script work with gcp_profile, '
        'make this script generate a TSV with a header in the first row. '
        'The first column of such TSV will be ignored since it is usually timestamp. '
        'Check monitoring_script in '
        'https://cromwell.readthedocs.io/en/stable/wf_options/Google/'
        '#google-pipelines-api-workflow-options '
        'for details.',
    )
    group_dep = parent_submit.add_argument_group(
        title='dependency resolver for all backends',
        description='Cloud-based backends (gc and aws) will only use Docker '
        'so that "--docker URI_FOR_DOCKER_IMG" must be specified '
        'in the command line argument or as a comment "#CAPER '
        'docker URI_FOR_DOCKER_IMG" or value for "workflow.meta.caper_docker"'
        'in a WDL file',
    )
    group_dep.add_argument(
        '--docker',
        nargs='?',
        const='',
        default=None,
        help='URI for Docker image (e.g. ubuntu:latest). '
        'This can also be used as a flag to use Docker image address '
        'defined in your WDL file as a comment ("#CAPER docker") or '
        'as "workflow.meta.caper_docker" in WDL.',
    )
    group_dep_local = parent_submit.add_argument_group(
        title='dependency resolver for local backend',
        description='Singularity is for local backend only. Other backends '
        '(gcp and aws) will use Docker only. '
        'Local backend defaults to not use any container-based methods. '
        'Use "--singularity" or "--docker" to use containers',
    )
    group_dep_local.add_argument(
        '--singularity',
        nargs='?',
        const='',
        default=None,
        help='URI or path for Singularity image '
        '(e.g. ~/.singularity/ubuntu-latest.simg, '
        'docker://ubuntu:latest, shub://vsoch/hello-world). '
        'This can also be used as a flag to use Docker image address '
        'defined in your WDL file as a comment ("#CAPER singularity") or '
        'as "workflow.meta.caper_singularity" in WDL.',
    )
    group_dep_local.add_argument(
        '--no-build-singularity',
        action='store_true',
        help='Do not build singularity image before running a workflow. ',
    )

    group_slurm = parent_submit.add_argument_group('SLURM arguments')
    group_slurm.add_argument('--slurm-partition', help='SLURM partition')
    group_slurm.add_argument('--slurm-account', help='SLURM account')
    group_slurm.add_argument(
        '--slurm-extra-param', help='SLURM extra parameters. Must be double-quoted'
    )

    group_sge = parent_submit.add_argument_group('SGE arguments')
    group_sge.add_argument(
        '--sge-pe', help='SGE parallel environment. Check with "qconf -spl"'
    )
    group_sge.add_argument('--sge-queue', help='SGE queue. Check with "qconf -sql"')
    group_sge.add_argument(
        '--sge-extra-param', help='SGE extra parameters. Must be double-quoted'
    )

    group_pbs = parent_submit.add_argument_group('PBS arguments')
    group_pbs.add_argument('--pbs-queue', help='PBS queue')
    group_pbs.add_argument(
        '--pbs-extra-param', help='PBS extra parameters. Must be double-quoted'
    )

    # server
    parent_server = argparse.ArgumentParser(add_help=False)
    parent_server.add_argument(
        '--java-heap-server',
        default=Cromwell.DEFAULT_JAVA_HEAP_CROMWELL_SERVER,
        help='Cromwell Java heap size for "server" mode (java -Xmx)',
    )
    parent_server.add_argument(
        '--disable-auto-write-metadata',
        action='store_true',
        help='Disable automatic retrieval/update/writing of metadata.json upon workflow/task status change.',
    )

    # run
    parent_run = argparse.ArgumentParser(add_help=False)
    parent_run.add_argument(
        '-m',
        '--metadata-output',
        help='An optional directory path to output metadata JSON file',
    )
    parent_run.add_argument(
        '--java-heap-run',
        default=Cromwell.DEFAULT_JAVA_HEAP_CROMWELL_RUN,
        help='Cromwell Java heap size for "run" mode (java -Xmx)',
    )

    # list, metadata, abort
    parent_search_wf = argparse.ArgumentParser(add_help=False)
    parent_search_wf.add_argument(
        'wf_id_or_label',
        nargs='*',
        help='List of workflow IDs to find matching workflows to '
        'commit a specified action (list, metadata and abort). '
        'Wildcards (* and ?) are allowed.',
    )

    # server, all client subcommands
    parent_server_client = argparse.ArgumentParser(add_help=False)
    parent_server_client.add_argument(
        '--port',
        type=int,
        default=Cromwell.DEFAULT_SERVER_PORT,
        help='Port for Caper server',
    )
    parent_server_client.add_argument(
        '--no-server-heartbeat',
        action='store_true',
        help='Disable server heartbeat file.',
    )
    parent_server_client.add_argument(
        '--server-heartbeat-file',
        default=ServerHeartbeat.DEFAULT_SERVER_HEARTBEAT_FILE,
        help='Heartbeat file for Caper clients to get IP and port of a server',
    )
    parent_server_client.add_argument(
        '--server-heartbeat-timeout',
        type=int,
        default=ServerHeartbeat.DEFAULT_HEARTBEAT_TIMEOUT_MS,
        help='Timeout for a heartbeat file in Milliseconds. '
        'A heartbeat file older than '
        'this interval will be ignored.',
    )

    parent_client = argparse.ArgumentParser(add_help=False)
    parent_client.add_argument(
        '--hostname',
        '--ip',
        default=CromwellRestAPI.DEFAULT_HOSTNAME,
        help='Hostname (or IP address) of Caper server.',
    )

    # list
    parent_list = argparse.ArgumentParser(add_help=False)
    parent_list.add_argument(
        '-f',
        '--format',
        default=DEFAULT_LIST_FORMAT,
        help='Comma-separated list of items to be shown for "list" '
        'subcommand. Any key name in workflow JSON from Cromwell '
        'server\'s response is allowed. '
        'Available keys are "id" (workflow ID), "status", "str_label", '
        '"name" (WDL/CWL name), "submission" (date/time), "start", '
        '"end" and "user". '
        '"str_label" is a special key for Caper. See help context '
        'of "--str-label" for details',
    )
    parent_list.add_argument(
        '--hide-result-before',
        help='Hide workflows submitted before this date/time. '
        'Use the same (or shorter) date/time format shown in '
        '"caper list". '
        'e.g. 2019-06-13, 2019-06-13T10:07',
    )
    parent_list.add_argument(
        '--show-subworkflow',
        action='store_true',
        help='Show subworkflows in "caper list". '
        'WARNING: If there are too many subworkflows, '
        'this can result in crash of Caper/Cromwell server ',
    )

    # troubleshoot/debug
    parent_troubleshoot = argparse.ArgumentParser(add_help=False)
    parent_troubleshoot.add_argument(
        '--show-completed-task',
        action='store_true',
        help='Show information about completed tasks.',
    )
    parent_troubleshoot.add_argument(
        '--show-stdout', action='store_true', help='Show STDOUT for failed tasks.'
    )

    # gcp_monitor
    parent_gcp_monitor = argparse.ArgumentParser(add_help=False)
    parent_gcp_monitor.add_argument(
        '--json-format',
        action='store_true',
        help='Prints out outputs in a JSON format.',
    )

    # gcp_res_analysis
    parent_gcp_res_analysis = argparse.ArgumentParser(add_help=False)
    parent_gcp_res_analysis.add_argument(
        '--in-file-vars-def-json',
        help='JSON file to define task name and input file variabless '
        'to be included in resource analysis. '
        'Key: task name, wild-cards (*, ?) are allowed. '
        'Value: list of input file var names. '
        'e.g. "atac.align*": ["fastqs_R1", "fastqs_R2"]. '
        'Once this file is defined, tasks not included in it will be ignored.',
    )
    parent_gcp_res_analysis.add_argument(
        '--reduce-in-file-vars',
        choices=[method.name for method in list(ResourceAnalysisReductionMethod)],
        default=ResourceAnalysisReductionMethod.sum.name,
        help='Reduce X matrix (resource data) into a vector. '
        'e.g. summing up all input file sizes. '
        'Reducing X will convert a multiple linear regression into a single linear regression. '
        'This is useful since single linear regression requires much less data '
        '(at least 2 for each task). '
        'Choose NONE to keep all input file variables '
        'without reduction in the analysis. '
        '2D Scatter plot (--plot-pdf) will not available for analysis without reduction. '
        'If NONE then make sure that number of datasets (per task) '
        '> number of input file variables in a task.',
    )
    parent_gcp_res_analysis.add_argument(
        '--target-resources',
        nargs='+',
        default=list(ResourceAnalysis.DEFAULT_TARGET_RESOURCES),
        help='Keys for resources in a JSON gcp_monitor outputs, '
        'which forms y vector for a linear problem. '
        'Analysis will be done separately for each key (resource metric). '
        'See help for gcp_monitor to find available resources. '
        'e.g. stats.max.disk, stats.mean.cpu_pct.',
    )
    parent_gcp_res_analysis.add_argument(
        '--plot-pdf',
        help='Local path for a 2D scatter plot PDF file. '
        'Scatter plot will not be available if --reduce-in-file-vars is none.',
    )

    # cleanup
    parent_cleanup = argparse.ArgumentParser(add_help=False)
    parent_cleanup.add_argument(
        '--delete',
        action='store_true',
        help='DELETE OUTPUTS. caper cleanup runs in a dry-run mode by default. ',
    )
    parent_cleanup.add_argument(
        '--num-threads',
        default=URIBase.DEFAULT_NUM_THREADS,
        type=int,
        help='Number of threads for cleaning up workflow\'s outputs. '
        'This is used for cloud backends only.',
    )

    # all subcommands
    p_init = subparser.add_parser(
        'init',
        help='Initialize Caper\'s configuration file. THIS WILL OVERWRITE ON '
        'A SPECIFIED(-c)/DEFAULT CONF FILE. e.g. ~/.caper/default.conf.',
        parents=[parent_all, parent_init],
    )
    p_run = subparser.add_parser(
        'run',
        help='Run a single workflow without server',
        parents=[parent_all, parent_submit, parent_run, parent_runner, parent_backend],
    )
    p_server = subparser.add_parser(
        'server',
        help='Run a Cromwell server',
        parents=[
            parent_all,
            parent_server_client,
            parent_server,
            parent_runner,
            parent_backend,
        ],
    )
    p_submit = subparser.add_parser(
        'submit',
        help='Submit a workflow to a Cromwell server',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_submit,
            parent_backend,
        ],
    )
    p_abort = subparser.add_parser(
        'abort',
        help='Abort running/pending workflows on a Cromwell server',
        parents=[parent_all, parent_server_client, parent_client, parent_search_wf],
    )
    p_unhold = subparser.add_parser(
        'unhold',
        help='Release hold of workflows on a Cromwell server',
        parents=[parent_all, parent_server_client, parent_client, parent_search_wf],
    )
    p_list = subparser.add_parser(
        'list',
        help='List running/pending workflows on a Cromwell server',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_list,
        ],
    )
    p_metadata = subparser.add_parser(
        'metadata',
        help='Retrieve metadata JSON for workflows from a Cromwell server',
        parents=[parent_all, parent_server_client, parent_client, parent_search_wf],
    )
    p_troubleshoot = subparser.add_parser(
        'troubleshoot',
        help='Troubleshoot workflow problems from metadata JSON file or '
        'workflow IDs',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_troubleshoot,
        ],
    )
    p_debug = subparser.add_parser(
        'debug',
        help='Identical to "troubleshoot"',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_troubleshoot,
        ],
    )
    p_gcp_monitor = subparser.add_parser(
        'gcp_monitor',
        help='Tabulate task\'s resource data collected on '
        'instances run on Google Cloud Compute. '
        'Use this for any workflows run with Caper>=1.2.0 on gcp backend. '
        'This is for gcp backend only.',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_gcp_monitor,
        ],
    )
    p_gcp_res_analysis = subparser.add_parser(
        'gcp_res_analysis',
        help='Linear resource analysis on monitoring data collected on '
        'instances run on Google Cloud Compute. This is for gcp backend only. '
        'Use this for any workflows run with Caper>=1.2.0 on gcp backend. '
        'Calculates coefficients/intercept for task\'s required resources '
        'based on input file size of a task. '
        'For each task it solves a linear regression problem of y=Xc + i1 + e where '
        'X is a matrix (m by n) of input file sizes and '
        'c is a coefficient vector (n by 1) and '
        'i is intercept and 1 is ones vector. '
        'y is a vector (m by 1) of resource taken and '
        'e is residual to be minimized. '
        'm is number of dataset and n is number of input file variables. '
        'Each resource metric will be solved separately. '
        'Refer to --target-resources for details about available resource metrics. '
        'Output will be a tuple of coefficient vector and intercept. ',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_gcp_res_analysis,
        ],
    )
    p_cleanup = subparser.add_parser(
        'cleanup',
        help='Cleanup outputs of workflows.',
        parents=[
            parent_all,
            parent_server_client,
            parent_client,
            parent_search_wf,
            parent_cleanup,
        ],
    )

    if conf_file is None:
        # partially parse args to get conf file from cmd line
        known_args, _ = parent_all.parse_known_args()
        conf_file = known_args.conf
    conf_file = os.path.expanduser(conf_file)

    subparsers = [
        p_init,
        p_run,
        p_server,
        p_submit,
        p_abort,
        p_unhold,
        p_list,
        p_metadata,
        p_troubleshoot,
        p_debug,
        p_gcp_monitor,
        p_gcp_res_analysis,
        p_cleanup,
    ]
    if os.path.exists(conf_file):
        conf_dict = update_parsers_defaults_with_conf(
            parsers=subparsers, conf_file=conf_file, conf_key_map=PARAM_KEY_NAME_CHANGE
        )
    else:
        conf_dict = None

    return parser, conf_dict



import fnmatch
import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict

import numpy as np
from matplotlib import pyplot
from matplotlib.backends.backend_pdf import PdfPages
from sklearn import linear_model

from .cromwell_metadata import CromwellMetadata, convert_type_np_to_py
from .dict_tool import flatten_dict

logger = logging.getLogger(__name__)


class ResourceAnalysis(ABC):
    """
    Class constants:
        DEFAULT_REDUCE_IN_FILE_VARS:
            Function to be used for reducing x vector.
            e.g. sum, min, max, ...
        DEFAULT_TARGET_RESOURCES:
            Keys to make y vector.
    """

    DEFAULT_REDUCE_IN_FILE_VARS = sum
    DEFAULT_TARGET_RESOURCES = ('stats.max.mem', 'stats.max.disk')

    def __init__(self):
        """Solves y = f(X) in a statistical way where
        X is a matrix vector of input file sizes (e.g. size_each([bam, bowtie2_index_tar, ...]))
        y is a vector of resources (e.g. [max_mem, max_disk, ...])
        """
        self._task_resources = []

    @property
    def task_resources(self):
        return self._task_resources

    def collect_resource_data(self, metadata_jsons):
        """Collect resource data from parsing metadata JSON files.

        self._task_resources is an extended (across all workflows) list of resource monitoring
        result from CromwellMetadata.gcp_monitor():
        [
            {
                'task_name': TASK1,
                'workflow_id': WORKFLOW_ID_OF_TASK1,
                ...
            },
            {
                'task_name': TASK2,
                'workflow_id': WORKFLOW_ID_OF_TASK2,
                ...
            }
        ]

        Args:
            metadata_jsons:
                List of metadata JSON file URIs or metadata JSONs
                or CromwellMetadata objects.
        """
        self._task_resources = []
        for metadata_json in metadata_jsons:
            self._task_resources.extend(CromwellMetadata(metadata_json).gcp_monitor())

    def analyze(
        self,
        in_file_vars=None,
        reduce_in_file_vars=DEFAULT_REDUCE_IN_FILE_VARS,
        target_resources=DEFAULT_TARGET_RESOURCES,
        plot_pdf=None,
    ):
        """Find and analyze all tasks.
        Run `self.collect_resource_data()` first to collect resource data before analysis.

        Args:
            in_file_vars:
                Dict { TASK_NAME: [IN_FILE_VAR1, ...] }.
                If not None, only tasks defined in it will be analyzed.
                See ResourceAnalysis.analyze_per_task.__doc__ for details.
            reduce_in_file_vars:
                Python function (e.g. sum, max) to reduce x matrix into a vector.
                See ResourceAnalysis.analyze_per_task.__doc__ for details.
            target_resources:
                Keys (in dot notation) to make vector y.
            plot_pdf:
                Local file name for a PDF plot.
        Returns:
            Results in a dict form: {
                TASK_NAME: {
                    'x': X_DATA,
                    'y': Y_DATA,
                    'coeffs': ANALYSIS_RESULT
                }
            }
        """
        plot_pp = None
        if plot_pdf:
            plot_pp = PdfPages(plot_pdf)

        result = {}

        if in_file_vars:
            all_tasks = in_file_vars.keys()
        else:
            all_tasks = list(set([task['task_name'] for task in self.task_resources]))

        for task_name in all_tasks:
            result[task_name] = self.analyze_task(
                task_name,
                in_file_vars=in_file_vars[task_name] if in_file_vars else None,
                reduce_in_file_vars=reduce_in_file_vars,
                target_resources=target_resources,
                plot_pp=plot_pp,
            )

        if plot_pdf:
            plot_pp.close()

        return result

    def analyze_task(
        self,
        task_name,
        in_file_vars=None,
        reduce_in_file_vars=DEFAULT_REDUCE_IN_FILE_VARS,
        target_resources=DEFAULT_TARGET_RESOURCES,
        plot_pp=None,
    ):
        """Does resource analysis on a task.
        Run `self.collect_resource_data()` first to collect resource data before analysis.
        Then you can such collected data for each task.

        To use a general solver for y = f(x),
        convert task's raw resouce data into (x_matrix, y_vector) form.

        Input file sizes are converted into one single matrix.
        Such x_matrix should look like:
            [
                [IN_FILE1_SIZE1, IN_FILE2_SIZE1, ...],
                [IN_FILE1_SIZE2, IN_FILE2_SIZE2, ...],
                ...
            ]
        If reduce_in_file_vars is defined (e.g. sum).
        Such x_matrix will be reduced to a vector.
            [
                [SUM_ALL_IN_FILES_SIZE1],
                [SUM_ALL_IN_FILES_SIZE2],
                ...
            ]

        Each resource metric in `target_resources` becomes a y_vec.
        For example, there are two kinds of default resource metrics.
        i.e. max_mem and max_disk.
        Therefore, y_vec_1 should look like:
            [
                STATS_MAX_MEM1,
                STATS_MAX_MEM2,
                ...
            ]
        y_vec_2 should look like:
            [
                STATS_MAX_DISK1,
                STATS_MAX_DISK2,
                ...
            ]
        They are two separate problems.

        Args:
            in_file_vars:
                List of input file vars. [IN_FILE_VAR1, ...].
                Matching input file will be included in x_matrix of analysis.
                See `input_file_sizes` in CromwellMetadata.gcp_monitor.__doc__ for details.
                If None or False, then all input files will be used to make x_matrix.
            reduce_in_file_vars:
                Python function to reduce x_vector. e.g. sum, max.
                If None, then no reduction of vector x.
                If you don't reduce x, make sure that you have enough number of data (>=dim(x)).
                Some solver can fail due to lack of data. i.e. dim(problem, solver) > dim(data).
                For example, even a linear solver will fail with only one dataset.
            target_resources:
                Keys (in dot notation) to make vector y. e.g. ('stats.max.mem', 'stats.max.disk').
                See CromwellMetadata.gcp_monitor.__doc__ to find available keys.
                One vector y for each key, which means one solving for each key.
            plot_pp:
                Matplotlib's PDF backend PdfPages object to write plots on multiple pages.
                (task/resource per page).

        Returns:
            coeffs:
                Analysis result.
                e.g. (coeffs, intercept) for linear regression.
        """
        result = {}

        in_file_vars_found = set()
        x_data = defaultdict(list)
        y_data = defaultdict(list)

        logger.info('Analyzing task={task}'.format(task=task_name))
        # first look at task's optional/empty input file vars across all workflows
        # e.g. SE (single-ended) pipeline runs does not have fastqs_R2
        # but we want to mix both SE/PE (paired-ended) data.
        # so need to look at all workflows to check if optional/empty var is
        # actully a file var.
        matched_task_resources = []
        for task in self.task_resources:
            if not fnmatch.fnmatchcase(task['task_name'], task_name):
                continue
            matched_task_resources.append(task)

            for in_file_var, in_file_size in task['input_file_sizes'].items():
                if in_file_vars and in_file_var not in in_file_vars:
                    continue

                if in_file_size:
                    in_file_vars_found.add(in_file_var)

        for task in matched_task_resources:
            # gather y_data
            found_y_data = False
            for res_metric, res_val in flatten_dict(task, reducer='.').items():
                if res_metric in target_resources and res_val:
                    y_data[res_metric].append(res_val)
                    found_y_data = True

            # it's possible that y_data doesn't exists
            # if a task is done immediately after initializing
            # even before the monitoring script runs
            # so if there is no y_data, then ignore x_data too.
            if not found_y_data:
                continue

            # gather x_data
            for in_file_var in in_file_vars_found:
                in_file_size = task['input_file_sizes'].get(in_file_var)
                if in_file_size:
                    x_data[in_file_var].append(sum(in_file_size))
                else:
                    x_data[in_file_var].append(0)

        if reduce_in_file_vars:
            key = '{reduce_name}({vars})'.format(
                reduce_name=reduce_in_file_vars.__name__,
                vars=','.join(sorted(x_data.keys())),
            )
            # transpose to reduce(sum by default) file sizes
            # over all in_file_vars
            tranposed = np.transpose([vec for vec in x_data.values()])
            reduced = [reduce_in_file_vars(vec) for vec in tranposed]
            x_data = {key: reduced}

        # tranpose it to make x matrix
        x_matrix = np.transpose([vec for vec in x_data.values()])

        result = {'x': x_data, 'y': y_data, 'coeffs': {}}
        for res_metric, y_vec in y_data.items():
            result['coeffs'][res_metric] = self._solve(
                x_matrix=x_matrix,
                y_vec=y_vec,
                plot_y_label=res_metric,
                plot_title=task_name,
                plot_pp=plot_pp,
            )

        # a bit hacky way to recursively convert numpy type into python type
        json_str = json.dumps(result, default=convert_type_np_to_py)
        return json.loads(json_str)

    @abstractmethod
    def _solve(self, x_matrix, y_vec, plot_y_label=None, plot_title=None, plot_pp=None):
        raise NotImplementedError


class LinearResourceAnalysis(ResourceAnalysis):
    def _solve(self, x_matrix, y_vec, plot_y_label=None, plot_title=None, plot_pp=None):
        """Solve y = A(X) with linear regression.
        Also make a scatter plot (for one-dimensional x_matrix only).
        Use `reduce_in_file_vars` in ResourceAnalysis.analyze()
        to reduce a matrix into a vector.

        Args:
            x_matrix:
                X Matrix.
            y_vec:
                y vector.
            plot_y_label:
                y label for plot.
            plot_title:
                Plot title.
            plot_pp:
                Matplotlib's PDF backend PdfPages object.
        Returns:
            Tuple of (coeffs, intercept).
        """
        x_matrix = np.array(x_matrix)

        try:
            model = linear_model.LinearRegression().fit(x_matrix, y_vec)

        except ValueError:
            logger.error(
                'Failed to solve due to type/dim mismatch? '
                'Too few data or invalid resource monitoring script? '
                'title: {title}, y_label: {y_label}, '
                'y_vec={y_vec}, x_matrix: {x_matrix}'.format(
                    title=plot_title,
                    y_label=plot_y_label,
                    y_vec=y_vec,
                    x_matrix=x_matrix,
                ),
                exc_info=True,
            )
            return

        if plot_pp:
            if x_matrix.shape[1] > 1:
                logger.warning(
                    'Cannot make a 2D scatter plot. dim(x_matrix) > 1. '
                    'Multi-dimensional analysis without reducing x matrix?'
                )
            else:
                x_vec = x_matrix[:, 0]
                # scatter plot with a fitting line
                pyplot.scatter(x_vec, y_vec, s=np.pi * 3, color=(0, 0, 0), alpha=0.5)
                pyplot.plot(x_vec, model.coef_ * x_vec + model.intercept_)
                pyplot.title(plot_title)
                pyplot.xlabel('input_file_size')
                pyplot.ylabel(plot_y_label)
                pyplot.savefig(plot_pp, format='pdf')
                pyplot.clf()

        return list(model.coef_), model.intercept_



from copy import deepcopy
from textwrap import dedent

from caper.dict_tool import merge_dict
from caper.hocon_string import HOCONString

INCLUDE_CROMWELL = 'include required(classpath("application"))'


def get_test_hocon_str():
    hocon_str = dedent(
        """\
        include required(classpath("application"))
        backend {
          default = "gcp"
          providers {
            Local {
              actor-factory = "cromwell.backend.impl.sfs.config.ConfigBackendLifecycleActorFactory"
              config {
                default-runtime-attributes {
                  docker = "ubuntu:latest"
                }
                root = "/mnt/data/scratch/leepc12/caper_out"
              }
            }
          }
        }"""
    )
    return hocon_str


def get_test_hocon_str2():
    hocon_str2 = dedent(
        """\
        include required(classpath("application"))
        backend {
          providers {
            gcp {
              actor-factory = "GOOGLE"
            }
          }
        }"""
    )
    return hocon_str2


def get_test_hocon_str_multiple_includes():
    return dedent(
        """\
        include required(classpath("application"))
        include required(file("application"))
        include required(url("application"))
        include required("application.conf")
        level1 {
          include file("/srv/test.conf")
          level2 {
            include url("http://ok.com/test.conf")
            level3 {
              include classpath("test")
              level4 {
                include "test.conf"
                level5 {
                  include "test.hocon"
                }
              }
            }
          }
        }"""
    )


def get_test_dict(with_include=False):
    base_dict = {
        'backend': {
            'default': 'gcp',
            'providers': {
                'Local': {
                    'actor-factory': 'cromwell.backend.impl.sfs.config.ConfigBackendLifecycleActorFactory',
                    'config': {
                        'default-runtime-attributes': {'docker': 'ubuntu:latest'},
                        'root': '/mnt/data/scratch/leepc12/caper_out',
                    },
                }
            },
        }
    }
    if with_include:
        base_dict[
            'HOCONSTRING_INCLUDE_ad5c3c187d5107c099f66681f1896c70'
        ] = 'include required(classpath("application"))'

    return base_dict


def get_test_dict2():
    """Without "include" lines.
    """
    return {'backend': {'providers': {'gcp': {'actor-factory': 'GOOGLE'}}}}


def get_test_multiple_includes(with_include=False):
    if with_include:
        return {
            "HOCONSTRING_INCLUDE_ad5c3c187d5107c099f66681f1896c70": "include required(classpath(\"application\"))",
            "HOCONSTRING_INCLUDE_61b86ce2e19939719a2e043b923774e4": "include required(file(\"application\"))",
            "HOCONSTRING_INCLUDE_543d042c69d8a730bc2b5785ac2f13c9": "include required(url(\"application\"))",
            "HOCONSTRING_INCLUDE_9456b859a44adad9a3d00ff3fcbbc5ae": "include required(\"application.conf\")",
            "level1": {
                "HOCONSTRING_INCLUDE_0714deb341d3d6291199d4738656c32b": "include file(\"/srv/test.conf\")",
                "level2": {
                    "HOCONSTRING_INCLUDE_91f31b362d72089d09f6245e912efb30": "include url(\"http://ok.com/test.conf\")",
                    "level3": {
                        "HOCONSTRING_INCLUDE_906d6e6eff885e840b705c2e7be3ba2d": "include classpath(\"test\")",
                        "level4": {
                            "HOCONSTRING_INCLUDE_c971be2dbb00ef0b44b9e4bf3c57f5cb": "include \"test.conf\"",
                            "level5": {
                                "HOCONSTRING_INCLUDE_44cb98470497b76dde0ab244c70870f0": "include \"test.hocon\""
                            },
                        },
                    },
                },
            },
        }
    else:
        return {'level1': {'level2': {'level3': {'level4': {'level5': {}}}}}}


def test_from_dict():
    ref_d = get_test_dict()
    hs = HOCONString.from_dict(ref_d, include=INCLUDE_CROMWELL)
    print(str(hs))
    print(get_test_hocon_str())
    assert str(hs) == get_test_hocon_str()


def test_to_dict():
    hs = HOCONString(get_test_hocon_str())
    assert hs.to_dict(with_include=False) == get_test_dict(with_include=False)
    assert hs.to_dict(with_include=True) == get_test_dict(with_include=True)

    hs = HOCONString(get_test_hocon_str_multiple_includes())
    assert hs.to_dict(with_include=False) == get_test_multiple_includes(
        with_include=False
    )
    assert hs.to_dict(with_include=True) == get_test_multiple_includes(
        with_include=True
    )


def test_merge():
    s1 = get_test_hocon_str()
    s2 = get_test_hocon_str2()
    s3 = get_test_hocon_str_multiple_includes()

    d1 = get_test_dict()
    d2 = get_test_dict2()
    d3 = get_test_multiple_includes(True)

    dm12 = deepcopy(d1)
    merge_dict(dm12, d2)
    dm32 = deepcopy(d3)
    merge_dict(dm32, d2)

    hs1 = HOCONString(s1)
    hs2 = HOCONString(s2)
    hs3 = HOCONString(s3)
    hsm12 = HOCONString.from_dict(dm12, include=INCLUDE_CROMWELL)
    hsm32 = HOCONString.from_dict(dm32)

    assert str(hsm12) == hs1.merge(hs2)
    assert str(hsm12) == hs1.merge(d2)
    assert str(hsm12) == hs1.merge(s2)

    assert str(hsm32) == hs3.merge(hs2)
    assert str(hsm32) == hs3.merge(d2)
    assert str(hsm32) == hs3.merge(s2)

    # merge with update
    # item 1 should be updated with merged
    hs1_original_str = str(hs1)
    assert str(hsm12) == hs1.merge(hs2, update=True)
    assert str(hs1) == str(hsm12)
    assert hs1_original_str != str(hs1)


def test_get_contents():
    s2 = get_test_hocon_str2()
    hs2 = HOCONString(s2)

    assert hs2.get_contents(with_include=True).strip() == s2
    assert (
        hs2.get_contents(with_include=False).strip()
        == s2.replace(INCLUDE_CROMWELL, '').strip()
    )



import json
import logging
import os
import shutil
import socket
import tempfile

from autouri import AbsPath, AutoURI

from .cromwell_metadata import CromwellMetadata
from .cromwell_workflow_monitor import CromwellWorkflowMonitor
from .nb_subproc_thread import NBSubprocThread, is_fileobj_open

logger = logging.getLogger(__name__)


class PortAlreadyInUseError(Exception):
    pass


class WomtoolValidationFailed(Exception):
    pass


def install_file(f, install_dir, label):
    """Install f locally on install_dir.
    If f is already local then skip it.
    """
    if AbsPath(f).is_valid:
        return AbsPath(f).uri
    logger.info('Installing {label}... {f}'.format(label=label, f=f))
    path = os.path.join(os.path.expanduser(install_dir), AutoURI(f).basename)
    return AutoURI(f).cp(path)


class Cromwell:
    """Wraps Cromwell/Womtool.
    """

    DEFAULT_CROMWELL = 'https://github.com/broadinstitute/cromwell/releases/download/59/cromwell-59.jar'
    DEFAULT_WOMTOOL = (
        'https://github.com/broadinstitute/cromwell/releases/download/59/womtool-59.jar'
    )
    DEFAULT_CROMWELL_INSTALL_DIR = '~/.caper/cromwell_jar'
    DEFAULT_WOMTOOL_INSTALL_DIR = '~/.caper/womtool_jar'
    DEFAULT_JAVA_HEAP_CROMWELL_SERVER = '10G'
    DEFAULT_JAVA_HEAP_CROMWELL_RUN = '4G'
    DEFAULT_JAVA_HEAP_WOMTOOL = '1G'
    DEFAULT_SERVER_PORT = 8000
    SERVER_STATUS_STARTED = 'server_started'
    LOCALHOST = 'localhost'

    def __init__(
        self,
        cromwell=DEFAULT_CROMWELL,
        womtool=DEFAULT_WOMTOOL,
        cromwell_install_dir=DEFAULT_CROMWELL_INSTALL_DIR,
        womtool_install_dir=DEFAULT_WOMTOOL_INSTALL_DIR,
    ):
        """
        Args:
            cromwell:
                Cromwell JAR path/URI/URL.
            womtool:
                Womtool JAR path/URI/URL.
            cromwell_install_dir:
                Local directory to install Cromwell JAR.
            womtool_install_dir:
                Local directory to install Womtool JAR.
        """
        self._cromwell = cromwell
        self._womtool = womtool

        if not AbsPath(cromwell_install_dir).is_valid:
            raise ValueError(
                'crommwell_install_dir is not a valid absolute '
                'path. {path}'.format(path=cromwell_install_dir)
            )
        self._cromwell_install_dir = cromwell_install_dir

        if not AbsPath(womtool_install_dir).is_valid:
            raise ValueError(
                'womtool_install_dir is not a valid absolute '
                'path. {path}'.format(path=womtool_install_dir)
            )
        self._womtool_install_dir = womtool_install_dir

    def validate(
        self,
        wdl,
        inputs=None,
        imports=None,
        cwd=None,
        java_heap_womtool=DEFAULT_JAVA_HEAP_WOMTOOL,
    ):
        """Validate WDL/inputs/imports using Womtool.

        Returns:
            valid:
                Validated or not.
        """
        self.install_womtool()

        wdl_file = AutoURI(wdl)
        if not wdl_file.exists:
            raise FileNotFoundError(
                'WDL file does not exist. wdl={wdl}'.format(wdl=wdl)
            )
        if inputs:
            if not AutoURI(inputs).exists:
                raise FileNotFoundError(
                    'Inputs JSON defined but does not exist. i={i}'.format(i=inputs)
                )

        with tempfile.TemporaryDirectory() as tmp_d:
            if imports:
                if not AutoURI(imports).exists:
                    raise FileNotFoundError(
                        'Imports file defined but does not exist. i={i}'.format(
                            i=imports
                        )
                    )
                wdl_ = os.path.join(tmp_d, wdl_file.basename)
                wdl_file.cp(wdl_)
                shutil.unpack_archive(imports, tmp_d)
            else:
                wdl_ = wdl_file.localize_on(tmp_d)

            cmd = [
                'java',
                '-Xmx{heap}'.format(heap=java_heap_womtool),
                '-jar',
                '-DLOG_LEVEL={lvl}'.format(lvl='INFO'),
                self._womtool,
                'validate',
                wdl_,
            ]
            if inputs:
                cmd += ['-i', AutoURI(inputs).localize_on(tmp_d)]

            logger.info('Validating WDL/inputs/imports with Womtool...')

            stderr = ''

            def on_stderr(s):
                nonlocal stderr
                stderr += s

            th = NBSubprocThread(cmd, cwd=tmp_d, on_stderr=on_stderr, quiet=True)
            th.start()
            th.join()

            if th.returncode:
                raise WomtoolValidationFailed(
                    'RC={rc}\nSTDERR={stderr}'.format(rc=th.returncode, stderr=stderr)
                )

            logger.info('Passed Womtool validation.')

    def run(
        self,
        wdl,
        inputs=None,
        options=None,
        imports=None,
        labels=None,
        metadata=None,
        backend_conf=None,
        backend=None,
        fileobj_stdout=None,
        fileobj_troubleshoot=None,
        java_heap_cromwell_run=DEFAULT_JAVA_HEAP_CROMWELL_RUN,
        java_heap_womtool=DEFAULT_JAVA_HEAP_WOMTOOL,
        work_dir=None,
        cwd=None,
        on_status_change=None,
        dry_run=False,
    ):
        """Run Cromwell run mode (java -jar cromwell.jar run).
        This is a non-blocking function which returns a NBSubprocThread object.
        So this function itself doesn't return anything.
        However, its NBSubprocThread object has a return value which is validated
        after the thread is done (joined).
        Such return value is metadata dict, which is a final output of Cromwell run.
        You can simply get it by thread.returnvalue after thread is done.

        Args:
            inputs:.
                input JSON file (-i).
            options:
                workflow options JSON file (-o).
            imports:
                imports.zip file (-p).
            labels:
                labels file (-l).
            metadata:
                output metadata JSON file (-m).
            backend_conf:
                backend.conf file (-Dconfig.file=).
                Default backend defined in this file will be used.
                If no default backend is defined then "Local" (Cromwell's default)
                backend will be used.
            fileobj_stdout:
                File-like object to print Cromwell's STDOUT.
            fileobj_troubleshoot:
                File-like object to write auto-troubleshooting result after
                workflow is done.
            java_heap_cromwell_run:
                Java heap (java -Xmx) for Cromwell run mode.
            java_heap_womtool=DEFAULT_JAVA_HEAP_WOMTOOL,
                Java heap (java -Xmx) for Womtool validation.
            work_dir:
                Temp directory to store Cromwell's output metadata JSON file.
                Cromwell will run on "cwd". Not on this work_dir.
            cwd:
                Current working directory to run Cromwell on.
                This will be finally passed to subprocess.Popen(cwd=).
            on_status_change:
                Not implemnted yet.
                Callback function called while polling.
                This function should take 5 args
                    workflow_id:
                        UUID of a workflow
                    workflow_new_status:
                        New status for a workflow. None if no change.
                    task_id:
                        Tuple (task_name, shard_idx) to identify workflow's task.
                    task_new_status:
                        New status for a task, None if no change.
                    metadata:
                        metadata (dict) of a workflow.
            dry_run:
                Dry run.
        Returns:
            th:
                Thread for Cromwell's run mode. None if dry_run.
                Notes:
                    Thread's return value (th.returnvalue)
                    is Cromwell's output metadata dict.
                    It is None if Cromwell subprocess itself didn't run,
                    If it ran but workflow failed then metadata dict is not None.
        """
        self.install_cromwell()

        if work_dir is None:
            work_dir = tempfile.mkdtemp(prefix='cromwell-run')

        # LOG_LEVEL must be >=INFO to catch workflow ID from STDOUT
        cmd = [
            'java',
            '-Xmx{}'.format(java_heap_cromwell_run),
            '-XX:ParallelGCThreads=1',
            '-jar',
            '-DLOG_LEVEL={lvl}'.format(lvl='INFO'),
            '-DLOG_MODE=standard',
        ]

        if backend_conf:
            cmd += ['-Dconfig.file={}'.format(backend_conf)]
        cmd += [self._cromwell, 'run', wdl]
        if inputs:
            cmd += ['-i', inputs]
        if options:
            cmd += ['-o', options]
        if labels:
            cmd += ['-l', labels]
        if imports:
            cmd += ['-p', imports]
        if metadata is None:
            metadata = os.path.join(
                work_dir, CromwellMetadata.DEFAULT_METADATA_BASENAME
            )
        cmd += ['-m', metadata]

        logger.debug('cmd: {cmd}'.format(cmd=' '.join(cmd)))
        if dry_run:
            return

        wm = CromwellWorkflowMonitor(on_status_change=on_status_change, is_server=False)

        def on_stdout(stdout):
            nonlocal wm
            nonlocal fileobj_stdout

            if is_fileobj_open(fileobj_stdout):
                fileobj_stdout.write(stdout)
                fileobj_stdout.flush()
            wm.update(stdout)

        def on_finish():
            nonlocal metadata
            nonlocal fileobj_troubleshoot

            if os.path.exists(metadata):
                json_contents = AutoURI(metadata).read()
                if json_contents:
                    metadata_dict = json.loads(json_contents)
                    cm = CromwellMetadata(metadata_dict)
                    cm.write_on_workflow_root()

                    if cm.workflow_status != 'Succeeded' and fileobj_troubleshoot:
                        # auto-troubleshoot on terminate if workflow is not successful
                        logger.info('Workflow failed. Auto-troubleshooting...')
                        fileobj_troubleshoot.write(cm.troubleshoot())

                    # to make it a return value of the thread after it is done (joined)
                    return metadata_dict

        th = NBSubprocThread(
            cmd,
            cwd=cwd,
            on_stdout=on_stdout,
            on_finish=on_finish,
            subprocess_name='Cromwell',
        )
        th.start()

        return th

    def server(
        self,
        server_port=DEFAULT_SERVER_PORT,
        server_hostname=None,
        server_heartbeat=None,
        backend_conf=None,
        fileobj_stdout=None,
        embed_subworkflow=False,
        java_heap_cromwell_server=DEFAULT_JAVA_HEAP_CROMWELL_SERVER,
        auto_write_metadata=True,
        on_server_start=None,
        on_status_change=None,
        cwd=None,
        dry_run=False,
    ):
        """Run Cromwell server mode (java -jar cromwell.jar server).
        This is a non-blocking function that returns a Thread object of Cromwell server.
        Howerver, this Thread object has a property status that indicates whether
        the server is started and ready to take submissions.
        Such condition is thread.status == True.

        Args:
            server_port:
                Server port.
            server_hostname:
                Server hostname. If defined then the heartbeat file will be written
                with this hostname instead of socket.gethostname().
            server_heartbeat:
                ServerHeartbeat object to write hostname/port pair into a heartbeat file.
                Then it will be later used by CaperClient to find hostname/port of
                this server.
            backend_conf:
                backend.conf file for Cromwell's Java parameter
                "-Dconfig.file=".
                Default backend defined in this file will be used.
                If no default backend is defined then "Local" (Cromwell's default)
                backend will be used.
            fileobj_stdout:
                File object to write Cromwell's STDOUT on.
            embed_subworkflow:
                This class basically stores/updates metadata.JSON file on
                each workflow's root directory whenever there is status change
                of workflow (or its tasks).
                This flag ensures that any subworkflow's metadata JSON will be
                embedded in main (this) workflow's metadata JSON.
                This is to mimic behavior of Cromwell run mode's -m parameter.
            java_heap_cromwell_server:
                Java heap (java -Xmx) for Cromwell server mode.
            auto_write_metadata:
                Automatic retrieval/writing of metadata.json upon workflow/task's status change.
            on_server_start:
                On server start.
            on_status_change:
                (Not implemented yet)
                Callback function called while polling.
                function should take 5 args
                    workflow_id:
                        UUID of a workflow
                    workflow_new_status:
                        New status for a workflow. None if no change.
                    task_id:
                        Tuple (task_name, shard_idx) to identify workflow's task.
                    task_new_status:
                        New status for a task, None if no change.
                    metadata:
                        metadata (dict) of a workflow.
            cwd:
                This will be finally passed to subprocess.Popen(cwd=).
            dry_run:
                Dry run.
        Returns:
            th:
                Thread for Cromwell's server mode.
                Returns None if dry_run.
        """
        self.install_cromwell()

        # check if port is open
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((Cromwell.LOCALHOST, server_port))
        if not result:
            raise PortAlreadyInUseError(
                'Server port {p} is already taken. '
                'Try with a different port'.format(p=server_port)
            )

        # LOG_LEVEL must be >=INFO to catch workflow ID from STDOUT
        cmd = [
            'java',
            '-Xmx{}'.format(java_heap_cromwell_server),
            '-XX:ParallelGCThreads=1',
            '-jar',
            '-DLOG_LEVEL={lvl}'.format(lvl='INFO'),
            '-DLOG_MODE=standard',
            '-Dwebservice.port={port}'.format(port=server_port),
        ]
        if backend_conf:
            cmd += ['-Dconfig.file={}'.format(backend_conf)]
            logger.debug(
                'backend_conf contents:\n{s}'.format(s=AutoURI(backend_conf).read())
            )

        cmd += [self._cromwell, 'server']

        logger.debug('cmd: {cmd}'.format(cmd=' '.join(cmd)))
        if dry_run:
            return

        wm = CromwellWorkflowMonitor(
            server_port=server_port,
            is_server=True,
            embed_subworkflow=embed_subworkflow,
            auto_write_metadata=auto_write_metadata,
            on_server_start=on_server_start,
            on_status_change=on_status_change,
        )

        def on_stdout(stdout):
            """Returns 'server_started' when server is ready to take submissions.
            Return value of this callback function is to update .status
            of an NBSubprocThread object.
            """
            nonlocal fileobj_stdout
            nonlocal wm
            nonlocal server_heartbeat

            if is_fileobj_open(fileobj_stdout):
                fileobj_stdout.write(stdout)
                fileobj_stdout.flush()

            wm.update(stdout)
            if wm.is_server_started():
                if server_heartbeat and not server_heartbeat.is_alive():
                    server_heartbeat.start(port=server_port, hostname=server_hostname)
                return 'server_started'

        def on_finish():
            nonlocal server_heartbeat

            if server_heartbeat:
                server_heartbeat.stop()

        th = NBSubprocThread(
            cmd,
            cwd=cwd,
            on_stdout=on_stdout,
            on_finish=on_finish,
            subprocess_name='Cromwell',
        )
        th.start()

        return th

    def install_cromwell(self):
        self._cromwell = install_file(
            self._cromwell, self._cromwell_install_dir, 'Cromwell JAR'
        )
        return self._cromwell

    def install_womtool(self):
        self._womtool = install_file(
            self._womtool, self._womtool_install_dir, 'Womtool JAR'
        )
        return self._womtool
